#include <so_5/all.hpp>

// Базовый класс для агентов примера.
class base : public so_5::agent_t {
protected:
   // Периодический сигнал, который агент должен обрабатывать
   // в каждом из своих состояний.
   struct tick final : public so_5::signal_t {};

   // Периодический сигнал для смены состояния агента.
   struct change_state final : public so_5::signal_t {};

   // Состояния, в которых может находится агент.
   state_t st_first{this, "first"},
         st_second{this, "second"},
         st_third{this, "third"};

   // Идентификаторы периодических сообщений. Мы должны их сохранить,
   // чтобы периодические сообщения не перестали приходить.
   so_5::timer_id_t tick_timer_;
   so_5::timer_id_t change_state_timer_;

public:
   base(context_t ctx) : so_5::agent_t{std::move(ctx)} {}

   // Общая часть инициализации всех агентов.
   virtual void so_define_agent() override {
      // Меняем начальное состояние.
      this >>= st_first;

      // Подписываемся на change_state для перехода в следующее состояние.
      st_first.event([this](mhood_t<change_state>) { this >>= st_second; });
      st_second.event([this](mhood_t<change_state>) { this >>= st_third; });
      st_third.event([this](mhood_t<change_state>) { this >>= st_first; });
   }

   // Начальные действия, одинаковые для всех агентов.
   virtual void so_evt_start() override {
      using namespace std::chrono_literals;

      // Заставляем агента менять состояние каждые 250ms.
      change_state_timer_ = so_5::send_periodic<change_state>(*this, 250ms, 250ms);

      // Заставляем приходить сообщение tick.
      tick_timer_ = so_5::send_periodic<tick>(*this, 0ms, 100ms);
   }
};

// Первый агент, который обрабатывает сигналы tick во всех своих состояниях.
class first_agent final : public base {
public:
   using base::base;

   virtual void so_define_agent() override {
      base::so_define_agent();

      // Добавляем недостающие подписки.
      so_subscribe_self()
         .in(st_first).in(st_second).in(st_third)
         .event([](mhood_t<tick>){});
   }
};

// Второй агент, который не подписывается на tick в st_second.
class second_agent final : public base {
public:
   using base::base;

   virtual void so_define_agent() override {
      base::so_define_agent();

      // Добавляем недостающие подписки.
      so_subscribe_self()
         .in(st_first).in(st_third)
         .event([](mhood_t<tick>){});
   }
};

// Агент, который будет время от времени "включать" и "выключать" поток
// интересующих нас trace-сообщений.
class trace_controller final : public so_5::agent_t {
   // Сигнал для включения/выключения.
   struct on_off final : public so_5::signal_t {};

   // Состояния, в которых может находится этот агент.
   state_t st_on{this}, st_off{this};

   // Таймер для сигнала on_off.
   so_5::timer_id_t timer_;

public:
   trace_controller(context_t ctx) : so_5::agent_t{std::move(ctx)} {
      st_off.event([this](mhood_t<on_off>) {
            // Меняем состояние.
            this >>= st_on;

            // Включаем трассировку тех сообщений, для который не
            // найден метод-обработчик у подписчика. Это говорит о том,
            // что агент не подписан в своем текущем состоянии.
            so_environment().change_message_delivery_tracer_filter(
               so_5::msg_tracing::make_filter(
                  [](const so_5::msg_tracing::trace_data_t & td) {
                     // Пытаемся получить информацию об обработчике сообщения.
                     const auto handler_ptr = td.event_handler_data_ptr();
                     if(handler_ptr) {
                        // Эта информация в trace-сообщении есть.
                        // Теперь можно посмотреть, найден ли сам обработчик.
                        if(nullptr == *handler_ptr)
                           // Обработчик не найден. Это именно та ситуация,
                           // которую мы ждем. Trace-сообщение можно разрешить.
                           return true;
                     }
                     // Во всех остальных случаях блокируем trace-сообщение.
                     return false;
                  }));
         });

      st_on.event([this](mhood_t<on_off>) {
            // Меняем состояние.
            this >>= st_off;

            // Выключаем трассировку.
            so_environment().change_message_delivery_tracer_filter(
               // Этот фильтр будет запрещать вообще все trace-сообщения.
               so_5::msg_tracing::make_disable_all_filter());
         });

      this >>= st_off;
   }

   virtual void so_evt_start() override {
      using namespace std::chrono_literals;
      timer_ = so_5::send_periodic<on_off>(*this, 0ms, 1500ms);
   }
};

int main() {
   // Запускаем SObjectizer с двумя агентами в рамках одной кооперации.
   so_5::launch([](so_5::environment_t & env) {
         env.introduce_coop([](so_5::coop_t & coop) {
            // В пример будут входить по одному агенту типа
            // first_agent и second_agent.
            coop.make_agent<first_agent>();
            coop.make_agent<second_agent>();

            // А так же агент, который будет периодически включать и
            // выключать поток trace-сообщений.
            coop.make_agent<trace_controller>();
         });
      },
      [](so_5::environment_params_t & params) {
         // Разрешаем трассировку механизма доставки сообщений.
         // Направляем трассировочные сообщения на стандартый поток вывода.
         params.message_delivery_tracer(so_5::msg_tracing::std_cout_tracer());
         // Блокируем все trace-сообщения до тех пор, пока это не потребуется.
         params.message_delivery_tracer_filter(
               so_5::msg_tracing::make_disable_all_filter());
      });
}

